var Iterator = function (iterableWrapper, yField) {
  this._iterator = iterableWrapper.iterator();
  Object.defineProperty(this, 'yField', {value: yField});
};
Iterator.prototype.next = function () {
  var hasNext = this._iterator.hasNext();
  return {
    done: !hasNext,
    value: hasNext ? wrap(this._iterator.next(), this.yField, this.enclosingEntity) : null
  }
};

var ImmutableIterableWW = function (iterableWrapper, yField, enclosingEntity, linkName) {
  Object.defineProperty(this, '_wrapped', {value: iterableWrapper});
  Object.defineProperty(this, 'yField', {value: yField});
  Object.defineProperty(this, 'enclosingEntity', {value: enclosingEntity});
  Object.defineProperty(this, 'linkName', {value: linkName});

  var def = function (obj, name) {
    var w = obj._wrapped;
    Object.defineProperty(obj, name, {
      value: function () {
        return new Iterator(w, yField);
      }
    });
  };
  def(this, 'entries');
  def(this, 'values');

  Object.defineProperty(this, 'size', {
    get: function () {
      return this._wrapped.size();
    }
  });
};
ImmutableIterableWW.prototype.forEach = function (visitor) {
  var iterator = this._wrapped.iterator();
  var index = 0;
  while (iterator.hasNext()) {
    visitor(wrap(iterator.next(), this.yField, this.enclosingEntity), index, this);
    index++;
  }
};
ImmutableIterableWW.prototype.first = function () {
  return wrap(this._wrapped.first(), this.yField, this.enclosingEntity);
};
ImmutableIterableWW.prototype.last = function () {
  return wrap(this._wrapped.last(), this.yField, this.enclosingEntity);
};
ImmutableIterableWW.prototype.isEmpty = function () {
  return this._wrapped.isEmpty();
};
ImmutableIterableWW.prototype.isNotEmpty = function () {
  return this._wrapped.isNotEmpty();
};
ImmutableIterableWW.prototype.get = function (index) {
  return wrap(this._wrapped.get(index), this.yField, this.enclosingEntity);
};
ImmutableIterableWW.prototype.has = function (element) {
  return this._wrapped.contains(unwrap(element, this.yField, this.enclosingEntity));
};
ImmutableIterableWW.prototype.find = function (visitor) {
  var iterator = this._wrapped.iterator();
  while (iterator.hasNext()) {
    var elem = iterator.next();
    var value =  wrap(elem, this.yField, this.enclosingEntity);
    var res = visitor(value);
    if (res) {
      return value;
    }
  }
};
ImmutableIterableWW.prototype['@@iterator'] = function () {
  return new Iterator(this._wrapped, this.yField);
};
Object.defineProperty(ImmutableIterableWW.prototype, "_canCheckHistory", {
  value: function () {
    if (!this.enclosingEntity) {
      throw 'Can not access field changes since this Set is not a member of an entity';
    }
  }
});
Object.defineProperty(ImmutableIterableWW.prototype, "isChanged", {
  get: function () {
    this._canCheckHistory();
    return this.enclosingEntity.isChanged(this.linkName || this.yField);
  },
  enumerable: true
});
Object.defineProperty(ImmutableIterableWW.prototype, "added", {
  get: function () {
    this._canCheckHistory();
    return this.enclosingEntity.getAdded(this.linkName || this.yField);
  },
  enumerable: true
});
Object.defineProperty(ImmutableIterableWW.prototype, "removed", {
  get: function () {
    this._canCheckHistory();
    return this.enclosingEntity.getRemoved(this.linkName || this.yField);
  },
  enumerable: true
});

function createMutableIterableWW(obj, yField, enclosingEntity, linkName) {
  var MutableIterableWW = function () {};
  MutableIterableWW.prototype = new ImmutableIterableWW(obj, yField, enclosingEntity, linkName);
  MutableIterableWW.prototype.add = function (element) {
    return this._wrapped.add(unwrap(element, yField));
  };
  MutableIterableWW.prototype.delete = function (element) {
    return this._wrapped.remove(unwrap(element, yField));
  };
  MutableIterableWW.prototype.clear = function () {
    return this._wrapped.clear();
  };
  return new MutableIterableWW();
}

var findYCustomField = function (name, project) {
  var cfs = $$fieldsAndLinksMeta.get(project._wrapped);
  for (var i = 0; i < cfs.length; i ++) { // we want to look for the name first
    if (cfs[i].getName() === name) {
      return cfs[i];
    }
  }
  for (var i = 0; i < cfs.length; i ++) { // and then for localized name
    if (cfs[i].getLocalizedName() === name) {
      return cfs[i];
    }
  }
};

var findYIssueLink = function (name) {
  var links = $$fieldsAndLinksMeta.getLinks();
  for (var i = 0; i < links.length; i ++) {
      var link = links[i];
      if (link.getInward() === name || link.getOutward() === name) {
      return link;
    }
  }
};

var findField = function (typeName, fieldName, project) {
  var field = yTypes.getByName(typeName).findField(fieldName);
  if (field) {
    return field;
  }
  if (typeName === 'Issue') {
    var yCustomField = findYCustomField(fieldName, project);
    if (yCustomField) {
      return yCustomField;
    }
    var yIssueLink = findYIssueLink(fieldName);
    if (yIssueLink) {
      return yIssueLink;
    }
  }
  throw 'No field with name ' + fieldName + ' found for type ' + typeName;
};

var wrap = function (obj, yField, enclosingEntity, linkName) {
  var wrapped = obj;
  if (obj instanceof Packages.jetbrains.youtrack.workflow.wrappers.EnumReference) {
    wrapped = obj.getValue();
  } else if (obj instanceof Packages.jetbrains.youtrack.workflow.wrappers.PersistentEntityWrapper) {
    wrapped = new entities[obj.getEntityType()](obj);
  } else if (obj instanceof Packages.jetbrains.youtrack.workflow.wrappers.MutableIterableWrapper) {
    wrapped = createMutableIterableWW(obj, yField, enclosingEntity, linkName);
  } else if (obj instanceof Packages.jetbrains.charisma.workflow.wrappers.WorkflowSequence) {
    wrapped = new ImmutableIterableWW(obj, yField, enclosingEntity, linkName);
  }
  return wrapped;
};

var isSprintsField = function (yField) {
  // TODO handle it in a more elegant way
  return yField && yField.getName() === 'sprints';
};

var unwrap = function(obj, yField) {
  var unwrapped = (!obj || !obj._wrapped) ? obj : obj._wrapped;
  if (isSprintsField(yField)) {
    return unwrapped;
  }
  if ((typeof unwrapped === 'string') && yField && (yField instanceof Packages.jetbrains.charisma.persistence.customfields.meta.YBundleCustomField)) {
    unwrapped = new Packages.jetbrains.youtrack.workflow.wrappers.EnumReference(unwrapped);
  }
  return unwrapped;
};

entities = {};

var unwrapArguments = function (argsArray) {
  if (!Array.isArray(argsArray)) {
    argsArray = Array.prototype.slice.call(argsArray, 0);
  }
  return argsArray.map(function (arg) {
    return (!arg || !arg._wrapped) ? arg : arg._wrapped;
  });
};

var getter = function (yField, name) {
  return function () {
    var result = safeCall(this._wrapped, 'get', [name ? name : yField.getName()], null);
    return wrap(result, yField, this.enclosingEntity || this, name);
  }
};

var setter = function (yField, name) {
  return function (value) {
    safeCall(this._wrapped, 'set', [name ? name : yField.getName(), unwrap(value, yField)], null);
  }
};

var method = function (name) {
  //TODO: add yField info for correct unwrapping
  return function () {
    return wrap(invoke(ctx.get(), this._wrapped, name, unwrapArguments(arguments)), null, this);
  }
};

var staticMethod = function (typeName, name) {
  return function () {
    return wrap(invokeStatic(ctx.get(), typeName, name, unwrapArguments(arguments)));
  }
};

var defineFunctions = function(target, applyTo, functions, enumerable) {
  Object.keys(functions).forEach(function (key) {
    Object.defineProperty(target, key, {
      value: function () {
        return functions[key].apply(applyTo, arguments);
      },
      enumerable: enumerable
    });
  });
};

var constructor = function (name) {
  var type = function () {
    var argsArray = Array.prototype.slice.call(arguments, 0);
    var _w;
    if (argsArray.length == 1 && argsArray[0] instanceof Packages.jetbrains.youtrack.workflow.wrappers.PersistentEntityWrapper) {
      _w = argsArray[0];
    } else {
      _w = invokeConstructor(ctx.get(), name, unwrapArguments(argsArray));
    }
    Object.defineProperty(this, '_wrapped', {value: _w});

    if (name === 'Issue') {
      var cfs = $$fieldsAndLinksMeta.get(this.project._wrapped);
      this.fields = {};
      Object.defineProperty(this.fields, '_wrapped', {value: this._wrapped});
      Object.defineProperty(this.fields, 'enclosingEntity', {value: this});
      for (var i = 0; i < cfs.length; i ++) {
        var cf = cfs[i];
        handleField(this, cf, cf.getName(), false);
        handleField(this.fields, cf, cf.getName(), true);
      }
      if ($$declareLinks !== null) {
        this.links = {};
        Object.defineProperty(this.links, '_wrapped', {value: this._wrapped});
        Object.defineProperty(this.links, 'enclosingEntity', {value: this});
        var linksMeta = $$fieldsAndLinksMeta.getLinks();
        for (var i = 0; i < linksMeta.length; i ++) {
          handleLink(this.links, linksMeta[i]);
        }
      }
    }
    var e = this;
    var fieldName = function (field) {
      if (!field) {
        throw "Invalid argument: " + field;
      }
      var ret = typeof field === 'string' ? field : field.name;
      if (!ret) {
        throw "Invalid argument: " + field;
      }
      return ret;
    };
    var entityFunctions = {
      required: function (field, message) {
        Packages.jetbrains.youtrack.workflow.model.Assertion.require(e._wrapped, fieldName(field), message);
      },
      becomes: function (field, expected) {
        return e._wrapped.becomes(fieldName(field), unwrap(expected, findField(name, fieldName(field), e.project)));
      },
      isChanged: function (field) {
        return e._wrapped.isChanged(fieldName(field));
      },
      oldValue: function (field) {
        return wrap(e._wrapped.getOldValue(fieldName(field)), findField(name, fieldName(field), e.project), e);
      },
      canBeReadBy: function (field, user) {
        return e._wrapped.canBeReadBy(fieldName(field), unwrap(user));
      },
      canBeWrittenBy: function (field, user) {
        return e._wrapped.canBeWrittenBy(fieldName(field), unwrap(user));
      }
    };
    defineFunctions(e, e, entityFunctions, true);
    var obsoleteEntityFunctions = { // these functions are obsolete since it's possible to call them on iterable wrappers
      getAdded: function (field) {
        return new ImmutableIterableWW(e._wrapped.added(fieldName(field)), findField(name, fieldName(field), e.project), e);
      },
      getRemoved: function (field) {
        return new ImmutableIterableWW(e._wrapped.removed(fieldName(field)), findField(name, fieldName(field), e.project), e);
      }
    };
    defineFunctions(e, e, obsoleteEntityFunctions, false);
    if (e.fields) {
      defineFunctions(e.fields, e, entityFunctions, true);
    }
  };
  Object.defineProperty(type, "$type", {value: name});
  return type;
};

var handleLink = function (target, yMember) {
  var defineLinkProperty = function (linksObject, yMember, name) {
    Object.defineProperty(linksObject, name, {
      enumerable: true,
      get: getter(yMember, name)
    });
  };
  defineLinkProperty(target, yMember, yMember.getOutward());
  if (yMember.getLocalizedOutward() && yMember.getLocalizedOutward() !== yMember.getOutward()) {
    defineLinkProperty(target, yMember, yMember.getLocalizedOutward());
  }
  if (yMember.getInward()) { // is directed
    defineLinkProperty(target, yMember, yMember.getInward());
    if (yMember.getLocalizedInward() && yMember.getLocalizedInward() !== yMember.getInward()) {
      defineLinkProperty(target, yMember, yMember.getLocalizedInward());
    }
  }
};

var handleField =  function (target, yMember, name, enumerable) {
  const readOnly = yMember.isReadOnly();
  Object.defineProperty(target, name, {
    enumerable: enumerable,
    get: getter(yMember, name),
    set: readOnly ? undefined : setter(yMember, name)
  });
  if (yMember instanceof Packages.jetbrains.charisma.persistence.customfields.meta.YCustomField) {
    // moreover, it's a custom field
    var localizedName = yMember.getLocalizedName();
    if (localizedName && localizedName !== name) {
      Object.defineProperty(target, localizedName, {
        enumerable: enumerable,
        get: getter(yMember, localizedName),
        set: setter(yMember, localizedName)
      });
    }
  }
};

var handleStaticMethod = function (target, typeName, memberName) {
  Object.defineProperty(target, memberName, {
    enumerable: true,
    value: staticMethod(typeName, memberName)
  });
};

var handleMethod = function (target, typeName, memberName) {
  Object.defineProperty(target, memberName, {
    enumerable: true,
    value: method(memberName)
  });
};

var handleProperty = function (target, yMember, propertyDescription, yClass) {
  var methodType = yMember.getMethodType();
  var isStatic = methodType.name() === 'STATIC';
  var getterName = propertyDescription.getterName;
  var setterName = propertyDescription.setterName;
  var getter = isStatic ? staticMethod(yClass.getName(), getterName) : method(getterName);
  var setter = isStatic ? staticMethod(yClass.getName(), setterName) : method(setterName);
  Object.defineProperty(target, propertyDescription.propertyName, {
    enumerable: true,
    get: getter,
    set: yClass.findMethod(setterName, 1, methodType) ? setter : undefined
  });
};

var toPropertyDescription = function (yMember) {
  var propertyName = yMember.getPropertyName();
  if (propertyName) {
    return {
      getterName: yMember.getName(),
      setterName: yMember.getCompliantSetterName(),
      propertyName: propertyName
    };
  }
};

var handleClass = function (yClass) {
  var yName = yClass.getName();
  if (entities[yName]) {
    return;
  }
  var entityType = constructor(yName);
  var superYClass = yClass.getSuperClass();
  if (superYClass) {
    var superYClassName = superYClass.getName();
    handleClass(superYClass);
    entityType.prototype = Object.create(entities[superYClassName].prototype);
    entityType.prototype.constructor = entities[superYClassName];
  }
  entities[yName] = entityType;

  var methodsToSkip = [];
  var staticMethodsToSkip = [];
  yClass.getDeclaredMembersArray().forEach(function (yMember) {
    if (!!yMember.isIgnoreInJs && yMember.isIgnoreInJs()) {
      return;
    }

    var memberName = yMember.getName();
    if (yMember instanceof Packages.jetbrains.charisma.persistence.customfields.meta.YField) {
      handleField(entityType.prototype, yMember, memberName, true);
    } else if (yMember.getMethodType().name() === 'CONSTRUCTOR') {
      // do nothing
    } else {
      var isStatic = yMember.getMethodType().name() === 'STATIC';
      var skip = isStatic ? staticMethodsToSkip : methodsToSkip;
      if (skip.indexOf(memberName) > -1) {
        return;
      }
      var target = isStatic ? entityType : entityType.prototype;
      var propertyDescription = toPropertyDescription(yMember);
      if (propertyDescription) {
        // this means a property should be defined instead of a method
        handleProperty(target, yMember, propertyDescription, yClass);
        skip.push(propertyDescription.setterName);
      } else {
        (isStatic ? handleStaticMethod : handleMethod)(target, yName, memberName)
      }
    }
  });
};
yTypes.get().forEach(handleClass);

// publish
Object.keys(entities).forEach(function (typeName) {
  exports[typeName] = entities[typeName];
});